"""Module defining the recorder class and configurations."""


from dataclasses import dataclass

import pyautogui
from pynput.keyboard import Controller, Key, Listener

from ..actions import Actions
from ..arkey import ARKey
from ..filehandler import FileHandler
from .state import ActionRecordingState


@dataclass
class RecorderConfig:
    """Configure the recorder class."""

    MOVE_KEY: Key = Key.ctrl_l
    CLICK_KEY: Key = Key.shift_l
    INPUT_RECORD_KEY: Key = "i"
    WAIT_KEY: str = "w"
    VARIABLE_PLACE_KEY: str = "v"
    END_ACTION_KEY: Key = Key.caps_lock
    EXIT_KEY: Key = Key.alt_l


class Recorder:
    """Record a set of actions."""

    def __init__(self, filename, verbose=0, config=None):
        """Initialize the recorder object."""
        self.filename = filename
        self.verbose = verbose
        self.actions = Actions(pyautogui.size())
        self.listener = None

        self.config = config or RecorderConfig()

        self._keyboard = Controller()
        self.state = ActionRecordingState(self)

        self.caps = 0

    def set_state(self, state):
        """State machine pattern's set_state method."""
        self.state = state

    def print(self, message, verbosity=1):
        """Print a message according to verbosity."""
        if self.verbose == verbosity:
            print(message)

    def _on_press(self, key):
        """Define the on_press method of the pynput listener."""
        self.state.key_down(ARKey(key))

    def _on_release(self, key):
        """Define the on_release method of the pynput listener."""
        self.state.key_pressed(ARKey(key))

    def start(self):
        """Start recording."""
        with Listener(
            on_press=self._on_press, on_release=self._on_release
        ) as self.listener:
            self.listener.join()

    def exit(self):
        """Exit the recording, saving the actions."""
        FileHandler.save(self)
        if self.caps % 2 == 1:
            keyboard = Controller()
            keyboard.press(Key.caps_lock)
            keyboard.release(Key.caps_lock)

        self.listener.stop()



"""Main script for arexecute application."""


import argparse
from ctypes import ArgumentError

from .core import executer, recorder

INSTRUCTIONS = """(->) Denotes press first one key, then the next
                  Alt                - Stop recording any time
             W -> any number         - Add waiting time of seconds equal to the number, floats allowed with a dot
                   I                 - Starts writing anything, commands allowed
               Caps Lock             - Stops any action (for example set the final wait time or stop clicking)
                  Ctrl               - Move mouse to current mouse position
             Shift n times           - Clicks n times in the last mouse position or current if there were no movements before
                   v                 - Adds a variable to be defined later
"""


def main():
    parser = argparse.ArgumentParser(
        description="Record/Execute keyboard and mouse actions.",
        prog="Recorder/Executer",
    )

    parser.add_argument(
        "-e",
        action="store",
        dest="execute",
        default=None,
        help="Runs the execution recursively until process killed or for a number of iterations if -i option is specificed. No effect on recording.",
    )
    parser.add_argument(
        "-r", action="store", dest="record", default=None, help="Runs recording mode."
    )
    parser.add_argument(
        "-i",
        nargs="?",
        const=-1,
        default=1,
        action="store",
        dest="iterations",
        type=int,
        help="Sets the execution mode on with i iterations (defaults to 1 iteration when number not specified).",
    )

    args = parser.parse_args()

    if not ((args.execute is None) != (args.record is None)):
        raise ArgumentError(
            "Please specify either the -e option or the -r option with a filename (specify only one)."
        )

    print("\n========================ACTION RECORD EXECUTE========================\n")

    if args.execute is not None:
        json_filename = args.execute
        execute = True

    else:
        json_filename = args.record
        execute = False

    iterations = args.iterations

    if execute:
        e = executer.Executer(json_filename, verbose=1)
        if iterations == -1:
            while True:
                e.start()

        else:
            for _ in range(iterations):
                e.start()

    else:
        print(INSTRUCTIONS)
        r = recorder.Recorder(json_filename, verbose=1)
        r.start()

    input("Process finished successfully, press enter to leave...\n")


if __name__ == "__main__":
    main()



"""Module defining the executer class and configuration."""


from dataclasses import dataclass

import pyautogui
from pynput.keyboard import Controller

from ..arkey import ARKey
from ..filehandler import FileHandler


@dataclass
class ExecuterConfig:
    """Dataclass defining the executer configurations."""

    MOVE_DURATION: float = 0.2
    CLICK_INTERVAL: float = 0.1
    WRITE_SPEED: float = 0.1


class Executer:
    """Execute a set of actions."""

    def __init__(self, filename, variables=None, verbose=0, config=None):
        """Initialize the executer."""
        self.filename = filename
        self.verbose = verbose

        self.current_position = None
        self.config = config or ExecuterConfig()

        self.actions = FileHandler.load(self.filename)
        self.translations = {v: k for k, v in self.actions.ACTION_TRANSLATION.items()}

        self.variables = variables
        self.verify_variables()

        self.size_conversion = [
            new_size / ref_size
            for new_size, ref_size in zip(pyautogui.size(), self.actions.screen_size)
        ]
        self._keyboard = Controller()

    def verify_variables(self):
        """Verify that all variables in actions are defined."""
        variable_count = [
            self.translations[action[0]] for action in self.actions.get_actions_list()
        ].count("variable")

        real_variable_count = 0 if self.variables is None else len(self.variables)

        if real_variable_count != variable_count:
            raise ValueError(
                "Specify the same amount of variables that are defined in the recorded action."
            )

    def start(self):
        """Start the execution."""
        for action in self.actions.get_actions_list():
            self.__getattribute__(self.translations[action[0]])(action[1])

    def move(self, pos):
        """Execute move action."""
        conv_pos = [
            pos * conversion for pos, conversion in zip(pos, self.size_conversion)
        ]
        self.current_position = conv_pos
        pyautogui.moveTo(conv_pos[0], conv_pos[1], duration=self.config.MOVE_DURATION)

    def click(self, clicks):
        """Execute click action."""
        pos = self.current_position or list(pyautogui.position())
        pyautogui.click(
            pos[0], pos[1], clicks=clicks, interval=self.config.CLICK_INTERVAL
        )

    def variable(self, index):
        """Execute insert variable action."""
        pyautogui.typewrite(self.variables[index], self.config.WRITE_SPEED)

    def type_input(self, inp):
        """Execute input typing action."""
        inp = [[ARKey(elem) for elem in inp_list] for inp_list in inp]

        for inp_list in inp:
            for elem in inp_list:
                self._keyboard.press(elem.key_for_action())

            for elem in inp_list:
                self._keyboard.release(elem.key_for_action())



"""Module that defines file handling elements."""


import json
from pathlib import Path

from .actions import Actions
from .arkey import ARKey


class CustomJSONEncoder(json.JSONEncoder):
    """JSON Encoder that supports ARKey."""

    def default(self, obj):
        if isinstance(obj, ARKey):
            return obj.to_json()

        return json.JSONEncoder.default(self, obj)


class JSONHandler:
    """File handler that uses JSON."""

    EXTENSION = ".json"

    @staticmethod
    def save(recorder, filepath):
        """Save the recorder actions to a file."""
        actions_dict = {
            "actions": recorder.actions.get_actions_list(),
            "screen_size": recorder.actions.screen_size,
        }

        with open(filepath, "w") as f:
            json.dump(actions_dict, f, cls=CustomJSONEncoder)

    @staticmethod
    def load(filepath):
        """Load actions from a file."""
        with open(filepath, "r") as f:
            actions_dict = json.load(f)

        actions = Actions(
            actions_dict["screen_size"], actions=actions_dict["actions"]
        )
        return actions


class FileHandler:
    """General file handler.

    Handlers save and load actions into a certain type of file.
    """

    HANDLER = JSONHandler

    @staticmethod
    def _preprocess(filename):
        """Preprocess the filename and convert it to a Path."""
        if not str(filename).endswith(FileHandler.HANDLER.EXTENSION):
            filename = str(filename) + FileHandler.HANDLER.EXTENSION

        return Path(str(filename))

    @staticmethod
    def save(recorder):
        """Save the recorder actions."""
        filepath = FileHandler._preprocess(recorder.filename)
        FileHandler.HANDLER.save(recorder, filepath)

    @staticmethod
    def load(filename):
        """Load a file with actions."""
        filepath = FileHandler._preprocess(filename)
        return FileHandler.HANDLER.load(filepath)



"""Module that defines a wrapper for keys defined in pynput."""


from pynput.keyboard import Key, KeyCode


class ARKey:
    """Wrapper for pynput Key class.

    This class provides an implementation to translate some key combinations
    that are not well defined in pynput, like ctrl and another key at the
    same time.
    """

    def __init__(self, key):
        """Initialize ARKey class."""
        self.isctrl = False
        self.ischar = False

        if isinstance(key, (Key, KeyCode)):
            self.isctrl = (key == Key.ctrl_l) or (key == Key.ctrl_r)
            self.key = key

        elif isinstance(key, int):
            try:
                self.key = Key(KeyCode.from_vk(key))

            except ValueError:
                self.key = KeyCode.from_vk(key)

        elif isinstance(key, str):
            self.key = KeyCode.from_char(key)

    def __eq__(self, other):
        """Overload equality operator."""
        return self.key == other.key

    def key_for_action(self):
        """Return the pynput key."""
        return self.key

    def processed(self):
        """Preprocess a key to compare in a normalized way.

        The preprocessing only includes lowercase transformation.
        """
        if isinstance(self.key, KeyCode):
            return self.key.char.lower()

        else:
            return self.key

    def to_json(self):
        """Transform a key object to a value for JSON formatting."""
        if isinstance(self.key, Key):
            value = self.key.value.vk

        elif isinstance(self.key, KeyCode):
            value = self.key.char

        else:
            value = self.key

        return value



from os import path

from setuptools import find_packages, setup

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="arexecute",
    version="1.0.1",
    description="Record and execute actions",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/jurrutiag/arexecute",
    author="Juan Urrutia",
    author_email="juan.urrutia.gandolfo@gmail.com",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    packages=find_packages(),
    install_requires=["pynput>=1.7.1", "pyautogui>=0.9.50"],
    extras_requires=["twine"],
    entry_points={"console_scripts": ["arexecute=arexecute.__main__:main"]},
)



"""Module that defines Action class."""


class Actions:
    """Defines a set of actions and methods to operate them.

    Actions are defined as things that are allowed to be done, these can be
    recorded and then executed.
    The ACTION_TRANSLATION constant defines the translation from a name to
    an abbreviation.
    """

    ACTION_TRANSLATION = {
        "move": "M",
        "click": "C",
        "variable": "V",
        "type_input": "I",
        "wait": "W",
    }

    def __init__(self, screen_size, actions=None):
        """Initialize Action object."""
        self.actions = actions or []
        self.screen_size = screen_size
        self.variables_index = 0

    def _add_action(self, action):
        """Add an action to the set."""
        self.actions.append(action)

    def get_actions_list(self):
        """Return the action set."""
        return self.actions

    def move(self, x, y):
        """Add a move action to an unnormalized x, y position."""
        action = (Actions.ACTION_TRANSLATION["move"], (x, y))
        self._add_action(action)

    def click(self, clicks):
        """Add a click action with n clicks."""
        action = (Actions.ACTION_TRANSLATION["click"], clicks)
        self._add_action(action)

    def variable(self):
        """Add a variable placeholder action."""
        action = (Actions.ACTION_TRANSLATION["variable"], self.variables_index)
        self.variables_index += 1
        self._add_action(action)

    def type_input(self, inputs):
        """Add a typing action."""
        action = (Actions.ACTION_TRANSLATION["type_input"], inputs)
        self._add_action(action)

    def wait(self, time):
        """Add a wait action for n seconds."""
        action = (Actions.ACTION_TRANSLATION["wait"], time)
        self._add_action(action)



"""Define the state machine for recording."""


import pyautogui
from pynput.keyboard import Key

from ..arkey import ARKey


class RecorderState:
    """Abstract class defining the state."""

    def __init__(self, machine):
        """Initialize the RecorderState."""
        self.machine = machine
        self.ended_action = False

    def key_pressed(self, key):
        """Reaction for when a key is pressed."""
        processed_key = key.processed()

        if processed_key == self.machine.config.MOVE_KEY:
            self.move_key_pressed()

        elif processed_key == self.machine.config.CLICK_KEY:
            self.clicking_key_pressed()

        elif processed_key == self.machine.config.INPUT_RECORD_KEY:
            self.input_recording_key_pressed()

        elif processed_key == self.machine.config.WAIT_KEY:
            self.waiting_key_pressed()

        elif processed_key == self.machine.config.VARIABLE_PLACE_KEY:
            self.variable_place_key_pressed()

        elif processed_key == self.machine.config.END_ACTION_KEY:
            self.end_action_key_pressed()

        elif processed_key == self.machine.config.EXIT_KEY:
            self.exit_key_pressed()

        self.any_key_pressed(key)

    def move_key_pressed(self):
        """Perform an action when move key is pressed."""
        pass

    def clicking_key_pressed(self):
        """Perform an action when click key is pressed."""
        pass

    def input_recording_key_pressed(self):
        """Perform an action when recording key is pressed."""
        pass

    def waiting_key_pressed(self):
        """Perform an action when waiting key is pressed."""
        pass

    def variable_place_key_pressed(self):
        """Perform an action when place variable key is pressed."""
        pass

    def end_action_key_pressed(self):
        """Perform an action when action key is pressed."""
        pass

    def exit_key_pressed(self):
        """Perform an action when exit key is pressed."""
        self.machine.set_state(ExitState(self.machine))

    def any_key_pressed(self, key):
        """Perform an action when any key is pressed."""
        pass

    def key_down(self, key):
        """Perform an action when a key is pressed down."""
        pass


class ActionRecordingState(RecorderState):
    """Default state for recording."""

    def __init__(self, *args, ending=False, **kwargs):
        """Initialize the default state."""
        super().__init__(*args, **kwargs)

        if ending:
            if self.machine.config.END_ACTION_KEY == Key.caps_lock:
                self.machine._keyboard.press(Key.caps_lock)
                self.machine._keyboard.release(Key.caps_lock)

        self.machine.print("- Listening for actions -")

    def move_key_pressed(self):
        # Coordinates are on image coordinates.
        x, y = list(pyautogui.position())
        self.machine.actions.move(x, y)
        self.machine.print(f"Moved to {x}, {y}")

    def clicking_key_pressed(self):
        self.machine.set_state(ClickingState(self.machine))

    def input_recording_key_pressed(self):
        self.machine.set_state(InputRecordingState(self.machine))

    def waiting_key_pressed(self):
        self.machine.set_state(WaitingTimeSetState(self.machine))

    def variable_place_key_pressed(self):
        self.machine.actions.variable()
        self.machine.print("Variable placed")


class ClickingState(RecorderState):
    """State activated when performing clicks."""

    def __init__(self, *args, **kwargs):
        """Initialize the clicking state."""
        super().__init__(*args, **kwargs)
        self.clicks = 1
        self.machine.print("- Listening for clicks (1 click total) -")

    def _record_clicks(self):
        """Record the clicks into the actions."""
        self.machine.actions.click(self.clicks)
        self.machine.print(f"{self.clicks} clicks recorded")

    def clicking_key_pressed(self):
        self.clicks += 1
        self.machine.print(f"- Listening for clicks ({self.clicks} clicks total) -")

    def end_action_key_pressed(self):
        self._record_clicks()
        self.machine.set_state(ActionRecordingState(self.machine, ending=True))


class InputRecordingState(RecorderState):
    """State activated when input is being recorded."""

    def __init__(self, *args, **kwargs):
        """Initialize the input recording state."""
        super().__init__(*args, **kwargs)
        self.inputs = []
        self.current_combination = []
        self.current_pressed = []
        self.machine.print("- Listening for input -")

    def key_down(self, key):
        if (
            key in self.current_pressed
            or key.processed() == self.machine.config.END_ACTION_KEY
        ):
            return

        if hasattr(key.key, "char") and any(k.isctrl for k in self.current_pressed):
            key = ARKey(chr(ord(key.key.char) + 64).lower())

        self.current_combination.append(key)
        self.current_pressed.append(key)

    def _record_input(self):
        """Record the input typing action."""
        self.machine.actions.type_input(self.inputs)
        self.machine.print("Recorded input")

    def end_action_key_pressed(self):
        self._record_input()
        self.machine.set_state(ActionRecordingState(self.machine, ending=True))

    def any_key_pressed(self, key):
        if key.processed() == self.machine.config.END_ACTION_KEY:
            return

        if hasattr(key.key, "char") and any(k.isctrl for k in self.current_pressed):
            key = ARKey(chr(ord(key.key.char) + 64).lower())

        if len(self.current_pressed) == 1:
            self.current_pressed = []

        if key in self.current_pressed:
            self.current_pressed.remove(key)

        if len(self.current_pressed) == 0:
            self.inputs.append(self.current_combination)
            self.current_combination = []

    def exit_key_pressed(self):
        pass


class ExitState(RecorderState):
    """State activated when the job is finished."""

    def __init__(self, *args, **kwargs):
        """Initialize the exit state."""
        super().__init__(*args, **kwargs)
        self.machine.exit()


class WaitingTimeSetState(RecorderState):
    """State activated when entering waiting time."""

    def __init__(self, *args, **kwargs):
        """Initialize the waiting time state."""
        super().__init__(*args, **kwargs)
        self.time = ""
        self.machine.print("- Listening for wait time, can be floating point -")

    def _record_waiting(self):
        """Record the waiting time into the actions."""
        self.machine.actions.wait(float(self.time))
        self.machine.print(f"Wait with {self.time} seconds recorded")

    def end_action_key_pressed(self):
        if self.time.count(".") > 1:
            self.machine.print(
                "Invalid time (can't have two dots), returning to action state"
            )

        else:
            self._record_waiting()

        self.machine.set_state(ActionRecordingState(self.machine, ending=True))

    def any_key_pressed(self, key):
        if key.processed() == self.machine.config.END_ACTION_KEY:
            return

        if key.processed() in "0123456789.":
            self.time += key.processed()
